package co.id.datacomsolusindo.ipphonebridge

import com.fasterxml.jackson.core.type.TypeReference
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.databind.ObjectWriter
import org.apache.commons.io.IOUtils
import org.apache.logging.log4j.LogManager
import org.springframework.context.annotation.Configuration
import org.springframework.core.Ordered
import org.springframework.core.annotation.Order
import org.springframework.core.io.ResourceLoader
import org.springframework.http.HttpHeaders
import org.springframework.http.HttpMethod
import org.springframework.http.HttpStatus
import org.springframework.messaging.handler.annotation.DestinationVariable
import org.springframework.messaging.handler.annotation.MessageMapping
import org.springframework.messaging.handler.annotation.SendTo
import org.springframework.messaging.simp.SimpMessagingTemplate
import org.springframework.stereotype.Component
import org.springframework.stereotype.Controller
import org.springframework.web.bind.annotation.GetMapping
import org.springframework.web.bind.annotation.PathVariable
import org.springframework.web.bind.annotation.PostMapping
import org.springframework.web.bind.annotation.RestController
import org.springframework.web.servlet.config.annotation.EnableWebMvc
import org.springframework.web.servlet.config.annotation.ResourceHandlerRegistry
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer
import org.springframework.web.servlet.resource.PathResourceResolver
import java.io.File
import java.io.IOException
import java.io.Serializable
import java.net.InetAddress
import java.time.Duration
import java.time.LocalDateTime
import java.util.*
import java.util.concurrent.TimeUnit
import javax.servlet.*
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import kotlin.concurrent.fixedRateTimer
import kotlin.streams.toList


@Component
@Order(Ordered.HIGHEST_PRECEDENCE)
class SimpleCORSFilter : Filter {
    @Throws(ServletException::class)
    override fun init(fc: FilterConfig?) {
    }

    @Throws(IOException::class, ServletException::class)
    override fun doFilter(
        req: ServletRequest, resp: ServletResponse,
        chain: FilterChain
    ) {
        val response = resp as HttpServletResponse
        val request = req as HttpServletRequest
        response.setHeader("Access-Control-Allow-Origin", "*")
        response.setHeader("Access-Control-Allow-Methods", "PATCH,PUT,POST,GET,OPTIONS,DELETE")
        response.setHeader("Access-Control-Max-Age", "3600")
        response.setHeader(
            "Access-Control-Allow-Headers",
            "x-requested-with, authorization, Content-Type, Authorization, credential, X-XSRF-TOKEN"
        )
        if ("OPTIONS".equals(request.method, ignoreCase = true)) {
            response.status = HttpServletResponse.SC_OK
        } else {
            chain.doFilter(req, resp)
        }
    }

    override fun destroy() {}
}

@Component
@Order(1)
class BridgeFilter(private val template: SimpMessagingTemplate, val resourceLoader: ResourceLoader) : Filter {
    override fun doFilter(request: ServletRequest, response: ServletResponse, chain: FilterChain) {
        val startTime = System.nanoTime()
        val req = request as HttpServletRequest
        val res = response as HttpServletResponse
        val splitPath = req.requestURI.split("/")
        val client = splitPath[1]
        val toPath = "/" + splitPath.takeLast(splitPath.size - 1).joinToString("/")
        val clientNumber = client.toIntOrNull()
        if (clientNumber != null && clientNumber > 0 && clientNumber <= 999) {
            val search = isResourceAvailable(toPath)
            if (search != null && !search.endsWith("assets/bootstrap.js")) {
                res.sendRedirect("/resources/$search")
            } else {
//                if (req.getHeader("User-Agent").contains("(dart:io)") || req.requestURL.toString().endsWith("api/license")) {
                val reqId = UUID.randomUUID().toString()
                try {
                    val headerMap = req.headerNames.toList().associateBy({ it }, { req.getHeader(it) })
                    val rb = RequestBuilder(reqId, toPath, HttpMethod.valueOf(req.method), headerMap.toMutableMap())

                    rb.queryString = req.queryString
                    if (req.contentType != null && req.contentType.toLowerCase().indexOf("multipart/form-data") > -1) {
                        rb.parts =
                            req.parts.map {
                                val fb64 = Base64.getEncoder().encodeToString(it.inputStream.readBytes())
                                FilePart(it.name, it.submittedFileName, fb64)
                            }.toMutableList()
                    } else {
                        val body = req.reader.lines().toList().joinToString(System.lineSeparator())
                        rb.body = body
                    }

                    Singletons.requestInstance[reqId] = RequestQue(reqId, rb, null)
                    val ow: ObjectWriter = ObjectMapper().writer().withDefaultPrettyPrinter()
                    val partStr = ow.writeValueAsString(rb).chunked(1500)


//                    val mid = UUID.randomUUID().toString()
//                    partStr.forEachIndexed(
//
//                    )
//                    val mList = mutableListOf<Message>()
//                    partStr.forEachIndexed { i, s ->
//                        mList.add(Message(mid, s, i + 1, partStr.size))
//                    }
//                    Singletons.sendQueue[mid] = mList.sortedBy { it.part }.toMutableList()
                    //send first
                    //template.convertAndSend("/topic/request/$client",Singletons.sendQueue[mid]!!.first())


//                    LogManager.getLogger(this.javaClass).info("part size ${partStr.size}")
//                    if (partStr.size > 1) {
                    val partialID = UUID.randomUUID().toString()
                    val st2 = System.nanoTime()

                    partStr.forEachIndexed { idx, it ->
                        template.convertAndSend(
                            "/topic/partial/$client",
                            PartialData(partialID, partStr.size, idx + 1, it, st2)
                        )
                    }


                    var i = 0
                    while (Singletons.responseQue[reqId] == null) {
                        TimeUnit.MILLISECONDS.sleep(100)
                        i++
                        if (i >= 600 * partStr.size) {
                            res.status = HttpStatus.REQUEST_TIMEOUT.value()
                            res.writer.write("{\"message\":\"Request timeout. Client not responding\"}")
                            res.flushBuffer()
                            Singletons.requestInstance.remove(reqId)
                            val endTime = System.nanoTime()
                            val duration =
                                (endTime - startTime).toDouble() / 1000000 //divide by 1000000 to get milliseconds.
                            LogManager.getLogger("client.$client")
                                .info("Request failed, No Response. Time: $duration  UA: " + req.getHeader("User-Agent") + " From: " + InetAddress.getLocalHost().hostAddress + " To: " + req.requestURL.toString() + "?" + req.queryString)

                            ClientHolder.addFailedRequest(client)
                            return
                        }
                    }

                    val resFromSocket = Singletons.responseQue[reqId]

                    res.status = resFromSocket!!.statusCode
                    resFromSocket.headers?.let {
                        it.entries
                            .filter { f -> resFromSocket.statusCode < 400 || (resFromSocket.statusCode >= 400 && f.key == "Content-Type") }
                            .forEach { en ->
                                res.setHeader(en.key, en.value.joinToString(","))
                            }
                    }

                    resFromSocket.body?.let {
                        res.outputStream.write(it)
                    }

                    res.flushBuffer()

                    Singletons.requestInstance.remove(reqId)
                    val endTime = System.nanoTime()

                    val duration = (endTime - startTime).toDouble() / 1000000 //divide by 1000000 to get milliseconds.
                    LogManager.getLogger("client.$client")
                        .info("Request Success. Time: $duration  UA: " + req.getHeader("User-Agent") + " From: " + req.remoteAddr + " To: " + req.requestURL.toString() + "?" + req.queryString)

                    ClientHolder.addSuccessRequest(client, duration)
                    return


                } catch (e: Exception) {
                    e.printStackTrace()
                    res.status = HttpStatus.REQUEST_TIMEOUT.value()
                    res.writer.write("{\"message\":\"Request timeout. Unknown error\"}")
                    res.flushBuffer()
                    Singletons.requestInstance.remove(reqId)
                    val endTime = System.nanoTime()
                    val duration = (endTime - startTime).toDouble() / 1000000 //divide by 1000000 to get milliseconds.
                    LogManager.getLogger("client.$client")
                        .info("Request failed, Unknown error. Time: $duration  UA: " + req.getHeader("User-Agent") + " From: " + InetAddress.getLocalHost().hostAddress + " To: " + req.requestURL.toString() + "?" + req.queryString)
                    ClientHolder.addFailedRequest(client)
                    return
                }
            }
        } else {
            chain.doFilter(request, response)
            return
        }
//        if (
//            client.startsWith("_")
//            || client.startsWith("resources")
//            || client.startsWith("clientStat")
//            || client.startsWith("actuator")
//        ) {
//            chain.doFilter(request, response)
//            return
//        } else {
//
//        }
    }

    fun isResourceAvailable(path: String, itr: Int = 1): String? {
        val splPath = path.split("/").takeLast(itr)
        val f = File("web")
        if (splPath.size >= itr && f.isDirectory) {
            val search = splPath.joinToString("/")
            return f.listFiles()?.filter { it.isDirectory }?.map { it.name }?.map {
                Pair(it, resourceLoader.getResource("file:web/$it/$search"))
            }?.filter { it.second.exists() && it.second.file.isFile }?.map { "${it.first}/$search" }?.firstOrNull()
                ?: isResourceAvailable(path, itr + 1)
        }
        return null
    }
}

class RequestQue(val id: String, val requestBuilder: RequestBuilder, var responseObj: Resp?)
class PartialData(val id: String, val total: Int, val idx: Int, val data: String, val startTime: Long)

class RequestBuilder(
    val id: String,
    val path: String,
    val method: HttpMethod,
    var headers: MutableMap<String, String>
) : Serializable {
    var body: String? = null
    var queryString: String? = null
    var parts: MutableList<FilePart>? = null
}

class FilePart(val name: String, val fileName: String, val data: String)

class Resp(val body: ByteArray?, val statusCode: Int, val headers: Map<String, Array<String>>? = null)

object Singletons {
    //    val sendQueue: ConcurrentMap<String, MessageHandler> by lazy { ConcurrentHashMap<String, MessageHandler>() }
    val responseQue: MutableMap<String, Resp> by lazy { mutableMapOf<String, Resp>() }
    val requestInstance: MutableMap<String, RequestQue> by lazy { mutableMapOf<String, RequestQue>() }
    val buildSocketChunkData: MutableMap<String, ChunkCollector> by lazy { mutableMapOf<String, ChunkCollector>() }
}

class SocketChunkData(
    val body: ByteArray?,
    val header: HttpHeaders?,
    val status: Int,
    val part: Int,
    val totalPart: Int
)

class ChunkCollector(private val id: String) {
    private val listSocketChunk = mutableListOf<SocketChunkData>()
    fun add(dt: SocketChunkData) {
        if (!listSocketChunk.any { it.part == dt.part }) {
            listSocketChunk.add(dt)
        }

        if (dt.part == dt.totalPart - 1) {
            val tStart = LocalDateTime.now()
            fixedRateTimer("timer-$id", false, 20, 100) {
                if (listSocketChunk.size == dt.totalPart) {
                    val bodyList = listSocketChunk.sortedBy { it.part }.mapNotNull { it.body }
                    var bd = ByteArray(0)
                    bodyList.forEach { bd += it }
                    Singletons.responseQue[id] = Resp(
                        bd,
                        dt.status,
                        dt.header?.entries?.associate { Pair(it.key, it.value.toTypedArray()) }
                    )
                    this.cancel()
                }
                if (Duration.between(tStart, LocalDateTime.now()).toMillis() > 6000) {
                    this.cancel()
                }
            }
        }
    }
}

@Controller
class SocketChecker {
//    @MessageMapping("/notification/{id}")
//    @SendTo("/topic/notification/{id}")
//    @Throws(Exception::class)
//    fun notification(@DestinationVariable("id") id: String, message: MutableMap<*, *>): MutableMap<*, *> {
//        return mutableMapOf(Pair("Hi", "I'am Ok"))
//    }

    @MessageMapping("/hi")
    @SendTo("/topic/healthCheck")
    @Throws(Exception::class)
    fun greeting(message: MutableMap<*, *>): MutableMap<*, *> {
        return mutableMapOf(Pair("Hi", "I'am Ok"))
    }

    @MessageMapping("/response/{id}")
    @SendTo("/topic/healthCheck")
    @Throws(Exception::class)
    fun responseMsg(message: SocketChunkData, @DestinationVariable("id") id: String): MutableMap<*, *> {
        try {
            LogManager.getLogger(this.javaClass).info("req sent ${message.part} of ${message.totalPart}")

            if (message.totalPart == 1) {
                Singletons.responseQue[id] = Resp(
                    message.body,
                    message.status,
                    message.header?.entries?.map { Pair(it.key, it.value.toTypedArray()) }?.toMap()
                )
            } else {
                var sc = Singletons.buildSocketChunkData[id]

                if (sc == null) {
                    Singletons.buildSocketChunkData[id] = ChunkCollector(id)
                    sc = Singletons.buildSocketChunkData[id]
                }
                sc?.add(message)
            }
        } catch (e: Exception) {
            e.printStackTrace()
        }

        return mutableMapOf(Pair("Hi", "I'am Ok"))
    }
}

@RestController
class BridgeRestController {
    @PostMapping("/_response/{id}")
    fun responseFromRest(@PathVariable("id") id: String, req: HttpServletRequest): String {
        val objectMapper = ObjectMapper()
        val typeRef: TypeReference<HashMap<String, Array<String>>> =
            object : TypeReference<HashMap<String, Array<String>>>() {}
        val body = IOUtils.toByteArray(req.getPart("body").inputStream)
        Singletons.responseQue[id] = Resp(
            body,
            String(IOUtils.toByteArray(req.getPart("status").inputStream)).toInt(),
            if (req.getPart("header") != null) {
                objectMapper.readValue(String(IOUtils.toByteArray(req.getPart("header").inputStream)), typeRef)
            } else {
                null
            }
        )

        return "{\"success\":true}"
    }

    @GetMapping("/_request/{id}")
    fun getResponseObj(@PathVariable("id") id: String): RequestBuilder {
        if (Singletons.requestInstance[id] == null) {
            throw Exception("not found")
        }
        return Singletons.requestInstance[id]!!.requestBuilder
    }

    @GetMapping("/clientStat")
    fun getClientData(): Map<String, Client> {
        return ClientHolder.get()
    }

    @GetMapping("/clientStat/{id}")
    fun getClientDataOne(@PathVariable("id") id: String): Client? {
        return ClientHolder.get()[id]
    }
}

@Configuration
@EnableWebMvc
class MvcConfig : WebMvcConfigurer {
    override fun addResourceHandlers(registry: ResourceHandlerRegistry) {
        registry
            .addResourceHandler("resources/**")
            .addResourceLocations("file:web/")
            .setCachePeriod(3600 * 24 * 30)
            .resourceChain(true)
            .addResolver(PathResourceResolver())
    }
}

